import importlib
from typing import Dict, Sequence

import torch

import tvl.backend
from tvl.backend import BackendFactory

# Explicitly set backends for particular device types.
_device_backends: Dict[str, BackendFactory] = {}
# Known backends. These will be searched if a device type does not have a backend factory
# set explicitly.
_known_backends = {
    'cpu': [
        'tvl_backends.fffr.FffrBackendFactory',     # PyPI package: tvl-backends-fffr
        'tvl_backends.pyav.PyAvBackendFactory',     # PyPI package: tvl-backends-pyav
        'tvl_backends.opencv.OpenCvBackendFactory', # PyPI package: tvl-backends-opencv
    ],
    'cuda': [
        'tvl_backends.fffr.FffrBackendFactory',     # PyPI package: tvl-backends-fffr
        'tvl_backends.nvdec.NvdecBackendFactory',   # PyPI package: tvl-backends-nvdec
    ],
}


def set_backend_factory(device_type, backend_factory):
    """Set the backend factory to be used for a particular device type."""
    _device_backends[device_type] = backend_factory


def _auto_set_backend_factory(device_type):
    """Attempt to automatically set the backend for `device_type` if not set already."""
    if device_type in _device_backends and _device_backends[device_type] is not None:
        return
    if device_type in _known_backends:
        for backend_name in _known_backends[device_type]:
            try:
                module_name, class_name = backend_name.rsplit('.', 1)
                module = importlib.import_module(module_name)
                set_backend_factory(device_type, getattr(module, class_name)())
                return
            except ImportError:
                pass


def get_backend_factory(device_type) -> BackendFactory:
    """Get the backend factory which will be used for a particular device type."""
    _auto_set_backend_factory(device_type)
    if device_type in _device_backends:
        return _device_backends[device_type]
    raise Exception(f'failed to find a backend factory for device type: {device_type}')


class VideoLoader:
    def __init__(self, filename, device, dtype=torch.float32, backend_opts=None):
        if isinstance(device, str):
            device = torch.device(device)
        self.backend = get_backend_factory(device.type).create(filename, device, dtype, backend_opts)

    def seek(self, time_secs):
        self.backend.seek(time_secs)

    def seek_to_frame(self, frame_index):
        self.backend.seek_to_frame(frame_index)

    def read_frame(self):
        return self.backend.read_frame()

    def read_frames(self, n):
        return self.backend.read_frames(n)

    @property
    def duration(self):
        return self.backend.duration

    @property
    def frame_rate(self):
        return self.backend.frame_rate

    @property
    def n_frames(self):
        return self.backend.n_frames

    @property
    def width(self):
        return self.backend.width

    @property
    def height(self):
        return self.backend.height

    def remaining_frames(self):
        """Iterate sequentially over remaining frames in the video."""
        more_frames = True
        while more_frames:
            try:
                yield self.read_frame()
            except EOFError:
                more_frames = False

    def read_all_frames(self):
        """Iterate over all frames in the video."""
        self.seek_to_frame(0)
        return self.remaining_frames()

    def select_frames(self, frame_indices):
        """Iterate over frames selected by frame index.

        Frames will be yielded in ascending order of frame index, regardless of the way
        `frame_indices` is ordered.

        Args:
            frame_indices (Sequence of int): Indices of frames to read.
        """
        return self.backend.select_frames(frame_indices)



"""Functions for transforming image data stored in PyTorch tensors.

This module is necessary since most of the transformations provided by the `torchvision` package
are applicable for PIL.Image images only. Since tvl may load video frames on the GPU, we want
to be able to take the computation to the data rather than moving the images to and from main
memory.

As an additional benefit, these functions are defined such that they also work in batched mode,
which is especially useful for videos.
"""

import math
from typing import Sequence

import torch
from torch.nn.functional import interpolate
from torchgeometry import warp_affine


def normalise(tensor, mean, stddev, inplace=False):
    """Normalise the image with channel-wise mean and standard deviation.

    Args:
        tensor (torch.Tensor): The image tensor to be normalised.
        mean (Sequence of float): Means for each channel.
        stddev (Sequence of float): Standard deviations for each channel.
        inplace (bool): Perform normalisation in-place.

    Returns:
        Tensor: The normalised image tensor.
    """
    mean = torch.as_tensor(mean, device=tensor.device)[..., :, None, None]
    stddev = torch.as_tensor(stddev, device=tensor.device)[..., :, None, None]

    if inplace:
        tensor.sub_(mean)
    else:
        tensor = tensor.sub(mean)

    tensor.div_(stddev)
    return tensor


def denormalise(tensor, mean, stddev, inplace=False):
    """Denormalise the image with channel-wise mean and standard deviation.

    Args:
        tensor (torch.Tensor): The image tensor to be denormalised.
        mean (Sequence of float): Means for each channel.
        stddev (Sequence of float): Standard deviations for each channel.
        inplace (bool): Perform denormalisation in-place.

    Returns:
        Tensor: The denormalised image tensor.
    """
    mean = torch.as_tensor(mean, device=tensor.device)[..., :, None, None]
    stddev = torch.as_tensor(stddev, device=tensor.device)[..., :, None, None]

    if inplace:
        return tensor.mul_(stddev).add_(mean)
    else:
        return torch.addcmul(mean, tensor, stddev)


def resize(tensor, size, mode='bilinear'):
    """Resize the image.

    Args:
        tensor (torch.Tensor): The image tensor to be resized.
        size (tuple of int): Size of the resized image (height, width).
        mode (str): The pixel sampling interpolation mode to be used.

    Returns:
        Tensor: The resized image tensor.
    """
    assert len(size) == 2
    out_shape = (*tensor.shape[:-2], *size)
    if tensor.ndimension() < 3:
        raise Exception('tensor must be at least 2D')
    elif tensor.ndimension() == 3:
        tensor = tensor.unsqueeze(0)
    elif tensor.ndimension() > 4:
        tensor = tensor.view(-1, *tensor.shape[-3:])
    align_corners = None
    if mode in {'linear', 'bilinear', 'trilinear'}:
        align_corners = False
    resized = interpolate(tensor, size=size, mode=mode, align_corners=align_corners)
    return resized.view(*out_shape)


def crop(tensor, t, l, h, w, padding_mode='constant', fill=0):
    """Crop the image, padding out-of-bounds regions.

    Args:
        tensor (torch.Tensor): The image tensor to be cropped.
        t (int): Top pixel coordinate.
        l (int): Left pixel coordinate.
        h (int): Height of the cropped image.
        w (int): Width of the cropped image.
        padding_mode (str): Padding mode (currently "constant" is the only valid option).
        fill (float): Fill value to use with constant padding.

    Returns:
        Tensor: The cropped image tensor.
    """
    # If the crop region is wholly within the image, simply narrow the tensor.
    if t >= 0 and l >= 0 and t + h <= tensor.size(-2) and l + w <= tensor.size(-1):
        return tensor[..., t:t+h, l:l+w]

    if padding_mode == 'constant':
        result = torch.full((*tensor.size()[:-2], h, w), fill,
                            device=tensor.device, dtype=tensor.dtype)
    else:
        raise Exception('crop only supports "constant" padding currently.')

    sx1 = l
    sy1 = t
    sx2 = l + w
    sy2 = t + h
    dx1 = 0
    dy1 = 0

    if sx1 < 0:
        dx1 = -sx1
        w += sx1
        sx1 = 0

    if sy1 < 0:
        dy1 = -sy1
        h += sy1
        sy1 = 0

    if sx2 >= tensor.size(-1):
        w -= sx2 - tensor.size(-1)

    if sy2 >= tensor.size(-2):
        h -= sy2 - tensor.size(-2)

    # Copy the in-bounds sub-area of the crop region into the result tensor.
    if h > 0 and w > 0:
        src = tensor.narrow(-2, sy1, h).narrow(-1, sx1, w)
        dst = result.narrow(-2, dy1, h).narrow(-1, dx1, w)
        dst.copy_(src)

    return result


def flip(tensor, horizontal=False, vertical=False):
    """Flip the image.

    Args:
        tensor (torch.Tensor): The image tensor to be flipped.
        horizontal: Flip horizontally.
        vertical: Flip vertically.

    Returns:
        Tensor: The flipped image tensor.
    """
    if horizontal == True:
        tensor = tensor.flip(-1)
    if vertical == True:
        tensor = tensor.flip(-2)
    return tensor


def affine(tensor, matrix):
    """Apply an affine transformation to the image.

    Args:
        tensor (torch.Tensor): The image tensor to be warped.
        matrix (torch.Tensor): The 2x3 affine transformation matrix.

    Returns:
        Tensor: The warped image.
    """
    is_unbatched = tensor.ndimension() == 3
    if is_unbatched:
        tensor = tensor.unsqueeze(0)
    warped = warp_affine(tensor, matrix, tensor.size()[-2:])
    if is_unbatched:
        warped = warped.squeeze(0)
    return warped


def rotate(tensor, degrees):
    """Rotate the image anti-clockwise about the centre.

    Args:
        tensor (torch.Tensor): The image tensor to be rotated.
        degrees (float): The angle through which to rotate.

    Returns:
        Tensor: The rotated image tensor.
    """
    rads = math.radians(degrees)
    h, w = tensor.size()[-2:]
    c = math.cos(rads)
    s = math.sin(rads)
    x = (w - 1) / 2
    y = (h - 1) / 2
    # Transformation matrix for clockwise rotation about the centre of the image.
    matrix = torch.tensor([[
        [ c, s, -c * x - s * y + x],
        [-s, c,  s * x - c * y + y],
    ]], dtype=torch.float32, device=tensor.device)
    return affine(tensor, matrix)


def fit(tensor, size, fit_mode='cover', resize_mode='bilinear', *, fill=0):
    """Fit the image within the given spatial dimensions.

    Args:
        tensor (torch.Tensor): The image tensor to be fit.
        size (tuple of int): Size of the output (height, width).
        fit_mode (str): 'fill', 'contain', or 'cover'. These behave in the same way as CSS's
                        `object-fit` property.
        fill (float): padding value (only applicable in 'contain' mode).

    Returns:
        Tensor: The resized image tensor.
    """
    # Modes are named after CSS object-fit values.
    assert fit_mode in {'fill', 'contain', 'cover'}

    if fit_mode == 'fill':
        return resize(tensor, size, mode=resize_mode)
    elif fit_mode == 'contain':
        ih, iw = tensor.shape[-2:]
        k = min(size[-1] / iw, size[-2] / ih)
        oh = round(k * ih)
        ow = round(k * iw)
        resized = resize(tensor, (oh, ow), mode=resize_mode)
        result = tensor.new_full((*tensor.size()[:-2], *size), fill)
        y_off = (size[-2] - oh) // 2
        x_off = (size[-1] - ow) // 2
        result[..., y_off:y_off + oh, x_off:x_off + ow] = resized
        return result
    elif fit_mode == 'cover':
        ih, iw = tensor.shape[-2:]
        k = max(size[-1] / iw, size[-2] / ih)
        oh = round(k * ih)
        ow = round(k * iw)
        resized = resize(tensor, (oh, ow), mode=resize_mode)
        y_trim = (oh - size[-2]) // 2
        x_trim = (ow - size[-1]) // 2
        result = crop(resized, y_trim, x_trim, size[-2], size[-1])
        return result

    raise Exception('This code should not be reached.')



from pathlib import Path

from setuptools import setup, find_packages

version = Path(__file__).absolute().parent.joinpath('VERSION').read_text('utf-8').strip()


setup(
    name='tvl',
    version=version,
    author='Aiden Nibali',
    license='Apache Software License 2.0',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'numpy',
        'torch',
        'torchgeometry>=0.1.2',
    ],
    extras_require={
        'FffrBackend': ['tvl-backends-fffr==' + version],
        'NvdecBackend': ['tvl-backends-nvdec==' + version],
        'PyAvBackend': ['tvl-backends-pyav==' + version],
        'OpenCvBackend': ['tvl-backends-opencv==' + version],
    },
)



from abc import ABC, abstractmethod

import torch


class Backend(ABC):
    def __init__(self, filename, device, dtype, seek_threshold):
        """Create a video-reading backend instance for a particular video file.

        Args:
            filename: Path to the video file.
            device:
            dtype:
            seek_threshold (int): Hint for predicting when seeking to the next target frame
                would be faster than reading and discarding intermediate frames. Setting this value
                close to the video's GOP size should be a reasonable choice.
        """
        self.filename = filename
        self.device = torch.device(device)
        if self.device.type == 'cuda':
            self.device = torch.device('cuda', torch.cuda.current_device())
        self.dtype = dtype
        self.seek_threshold = seek_threshold

    @property
    @abstractmethod
    def duration(self):
        """The duration of the video (in seconds)."""

    @property
    @abstractmethod
    def frame_rate(self):
        """The frame rate of the video (in frames per second)."""

    @property
    @abstractmethod
    def n_frames(self):
        """The number of frames in the video."""

    @property
    @abstractmethod
    def width(self):
        """The width of the image."""

    @property
    @abstractmethod
    def height(self):
        """The height of the image."""

    @abstractmethod
    def seek(self, time_secs):
        """Seek to the specified time in the video file."""

    def seek_to_frame(self, frame_index):
        self.seek(frame_index / self.frame_rate)

    @abstractmethod
    def read_frame(self):
        """Read a single video frame as an RGB PyTorch tensor."""

    def read_frames(self, n):
        return [self.read_frame() for _ in range(n)]

    def select_frames(self, frame_indices):
        # We will be loading unique frames in ascending index order.
        sorted_frame_indices = list(sorted(set(frame_indices)))

        pos = -(self.seek_threshold + 1)
        seq_len = 0
        seq_keepers = []
        for frame_index in sorted_frame_indices:
            if frame_index - pos > self.seek_threshold:
                # Read previous sequence
                if seq_len > 0:
                    frames = self.read_frames(seq_len)
                    for i in seq_keepers:
                        yield frames[i]
                    seq_keepers.clear()
                    seq_len = 0
                # Skip to desired location by seeking.
                self.seek_to_frame(frame_index)
            else:
                # Skip to desired location by reading and discarding intermediate frames.
                while pos < frame_index:
                    seq_len += 1
                    pos += 1
            # Read the frame that we care about.
            seq_keepers.append(seq_len)
            seq_len += 1
            pos = frame_index + 1
        if seq_len > 0:
            frames = self.read_frames(seq_len)
            for i in seq_keepers:
                yield frames[i]


class BackendFactory(ABC):
    @abstractmethod
    def create(self, filename, device, dtype, backend_opts=None) -> Backend:
        pass



from concurrent.futures import Executor

from torch.utils.data.dataloader import default_collate
from torch.utils.data.sampler import Sampler, BatchSampler, RandomSampler, SequentialSampler


class BatchDataIter:
    def __init__(self, loader):
        self.loader = loader
        self.batch_iter = iter(loader.batch_sampler)
        self.batch_buffer = []
        self.max_buffer_len = 2

    def _prepare_future_batches(self, batch_iter):
        while len(self.batch_buffer) < self.max_buffer_len:
            try:
                batch_indices = next(batch_iter)
            except StopIteration:
                break
            future_batch = [self.loader.dataset[index] for index in batch_indices]
            self.batch_buffer.append(future_batch)

    def __next__(self):
        self._prepare_future_batches(self.batch_iter)
        if len(self.batch_buffer) == 0:
            raise StopIteration()
        future_batch = self.batch_buffer.pop(0)
        batch = [example.result() for example in future_batch]
        self._prepare_future_batches(self.batch_iter)
        return self.loader.collate(batch)

    def __len__(self):
        return len(self.loader)


class BatchDataLoader:
    def __init__(self, dataset, *, batch_size=1, shuffle=False, sampler=None, batch_sampler=None,
                 collate_fn=default_collate, drop_last=False):
        """Loads batches of data from an asynchronous dataset.

        Args:
            dataset (AsyncDataset):
            batch_size (int):
            shuffle (bool):
            sampler (Sampler):
            batch_sampler (Sampler):
            collate_fn (function):
            drop_last (bool):
        """
        self.dataset = dataset
        self.batch_size = batch_size
        self.collate = collate_fn

        if sampler is not None and shuffle:
            raise ValueError('sampler option is mutually exclusive with shuffle')

        if batch_sampler is not None:
            if batch_size > 1 or shuffle or sampler is not None or drop_last:
                raise ValueError('batch_sampler option is mutually exclusive with batch_size, '
                                 'shuffle, sampler, and drop_last')
        else:
            if sampler is None:
                if shuffle:
                    sampler = RandomSampler(dataset)
                else:
                    sampler = SequentialSampler(dataset)
            batch_sampler = BatchSampler(sampler, batch_size, drop_last)

        self.sampler = sampler
        self.batch_sampler = batch_sampler

    def __iter__(self):
        return BatchDataIter(self)

    def __len__(self):
        return len(self.batch_sampler)


class AsyncDataset:
    def __init__(self, dataset, executor):
        """Wraps a synchronous dataset to enable asynchronous access to examples.

        Args:
            dataset (Dataset): the synchronous dataset to wrap
            executor (Executor): the executor supplying a pool of workers
        """
        self.dataset = dataset
        self.executor = executor

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        return self.executor.submit(self.dataset.__getitem__, index)
